/********************************************************************/
/*                                                                  */
/*      Provides rhe following code:                                */
/*        IRQ Handling                                              */
/*        MMR definitions                                           */
/*        newlib bindings                                           */
/*                                                                  */
/*		(C) Tilo Lutz, TiloLutz@gmx.de								*/
/*																	*/
/*		09.07.2008	Initial release                                 */
/*																	*/
/*																	*/
/********************************************************************/

#include "aduc.h"

void (*UNDEF)(void);		/* 	Function Pointers for Interupts 	*/
void (*SWI)(void);
void (*PAbt)(void);
void (*DAbt)(void);
void (*IRQ)(void);
void (*FIQ)(void);
void	UNDEF_Routine(void) __attribute__ ((interrupt ("UNDEF")));
void	SWI_Routine(void) __attribute__ ((interrupt ("SWI")));
void	PAbt_Routine(void) __attribute__ ((interrupt ("PABORT")));
void	DAbt_Routine(void)  __attribute__ ((interrupt ("DABORT")));
void	IRQ_Routine(void) __attribute__ ((interrupt ("IRQ")));
void	FIQ_Routine(void) __attribute__ ((interrupt ("FIQ")));


void UNDEF_Routine(void) {
	if (*UNDEF !=0x00) {
		UNDEF();
	}
}

void SWI_Routine(void) {
	if (*SWI !=0x00) {
		SWI();
	}
}

void PAbt_Routine(void) {
	if (*PAbt !=0x00) {
		PAbt();
	}
}

void DAbt_Routine(void) {
	if (*DAbt !=0x00) {
		DAbt();
	}
}

void IRQ_Routine(void) {
	if (*IRQ !=0x00) {
		IRQ();
	}
}

void FIQ_Routine(void) {
	if (*FIQ !=0x00) {
		FIQ();
	}
}


#ifdef USENEWLIB
	char (*STDIN)(void);
	void (*STDOUT)(char ptr);
	
#include <stdlib.h>
#include <reent.h>
#include <sys/stat.h>
	

_ssize_t _read_r(
    struct _reent *r, 
    int file, 
    void *ptr, 
    size_t len) {

	unsigned char *p;
	p = ptr;
	
	if (*STDIN !=0x00) {
		*p = STDIN();
		return (_ssize_t)1;
	}
	else { /* Pointer is not set ergo no data can be written */
		return (_ssize_t)0;
	}
}

_ssize_t _write_r (
    struct _reent *r, 
    int file, 
    const void *ptr, 
    size_t len) {

	size_t todo;
	const unsigned char *p;

	todo = len;
	p = ptr;

	if (*STDOUT !=0x00) {
		for( ; todo != 0; todo--) {
			if ( *p == '\n' ) {
				STDOUT('\r');
			}
			STDOUT(*p++);
	    }
		return (_ssize_t)len;			/* Number of bytes written.	*/
	}
	else { /* Pointer is not set ergo no data can be written */
		return (_ssize_t)0;
	}
}

int _close_r(

    struct _reent *r, 

    int file)

{

	return 0;

}



_off_t _lseek_r(
    struct _reent *r, 
    int file, 
    _off_t ptr, 
    int dir) {
	return (_off_t)0;	/*  Always indicate we are at file beginning.  */
}

int _fstat_r(
    struct _reent *r, 
    int file, 
    struct stat *st) {
	/*  Always set as character device.				*/
	st->st_mode = S_IFCHR;
	/* assigned to strong type with implicit 	*/
	/* signed/unsigned conversion.  Required by 	*/
	/* newlib.					*/
	return 0;
}

int isatty(int file); /* avoid warning */

int isatty(int file) {
	return 1;
}

#if 0
static void _exit (int n) {

label:  goto label; /* endless loop */
}
#endif 


/**** Locally used variables. ****/
// mt: "cleaner": extern char* end;
extern char end[];              /*  end is set in the linker command 	*/
				/* file and is the end of statically 	*/
				/* allocated data (thus start of heap).	*/
static char *heap_ptr;		/* Points to current end of the heap.	*/

/************************** _sbrk_r *************************************
 * Support function. Adjusts end of heap to provide more memory to
 * memory allocator. Simple and dumb with no sanity checks.
 *  struct _reent *r -- re-entrancy structure, used by newlib to
 *                      support multiple threads of operation.
 *  ptrdiff_t nbytes -- number of bytes to add.
 *                      Returns pointer to start of new heap area.
 *
 *  Note:  This implementation is not thread safe (despite taking a
 *         _reent structure as a parameter).
 *         Since _s_r is not used in the current implementation, 
 *         the following messages must be suppressed.
 */

void * _sbrk_r(
    struct _reent *_s_r, 
    ptrdiff_t nbytes)
{
	char  *base;		/*  errno should be set to  ENOMEM on error  */
	if (!heap_ptr) {	/*  Initialize if first time through.  */
		heap_ptr = end;
	}
	base = heap_ptr;	/*  Point to end of heap.  */
	heap_ptr += nbytes;	/*  Increase heap.  */
	return base;		/*  Return pointer to start of new heap area.  */
}




#endif
