/* ****************************************************************************************************** */
/*   											crt.s	                                                  */
/*                                                                                                        */
/*                       Assembly Language Startup Code for ADI ADUC7000                                  */
/*                                                                                                    	  */
/*                                                                                           			  */
/*                                                                                                        */
/*                                                                                                        */
/* Author:  James P Lynch      September 2, 2006                                                          */
/* Modified:  Tilo Lutz, TiloLutz@gmx.de      13.01.2008                                                  */
/* This Startup File is now for audc7000 mcus                                                             */
/* ****************************************************************************************************** */

/*
   Define required MMR Base value and offsets
*/

.equ    MMR_BASE_0,         0xFFFF0000
.equ    POWKEY1_OFFSET,     0x0404
.equ    POWCON_OFFSET,      0x0408
.equ    POWKEY2_OFFSET,     0x040C

/*
// <e> PLL Setup
//   <o1.0..2>  CD: PLL Multiplier Selection
//               <0-7>
//               <i> CD Value
// </e>
*/
	    .equ    PLL_SETUP,      1
        .equ    PLLCFG_Val,     0x00000000

/* Block Startup? */
		.equ	MCUNOSTART,			0


/* Stack Sizes */
.set  UND_STACK_SIZE, 0x00000010		/* stack for "undefined instruction" interrupts is 16 bytes 	 */
.set  ABT_STACK_SIZE, 0x00000010		/* stack for "abort" interrupts is 16 bytes                 	 */
.set  FIQ_STACK_SIZE, 0x00000080		/* stack for "FIQ" interrupts  is 128 bytes     				 */
.set  IRQ_STACK_SIZE, 0X00000080		/* stack for "IRQ" normal interrupts is 128 bytes  				 */
.set  SVC_STACK_SIZE, 0x00000080		/* stack for "SVC" supervisor mode is 128 bytes  				 */

/* Standard definitions of Mode bits and Interrupt (I & F) flags in PSRs (program status registers) 	 */
.set  MODE_USR, 0x10            		/* Normal User Mode 											 */
.set  MODE_FIQ, 0x11            		/* FIQ Processing Fast Interrupts Mode 							 */
.set  MODE_IRQ, 0x12            		/* IRQ Processing Standard Interrupts Mode 						 */
.set  MODE_SVC, 0x13            		/* Supervisor Processing Software Interrupts Mode 				 */
.set  MODE_ABT, 0x17            		/* Abort Processing memory Faults Mode 							 */
.set  MODE_UND, 0x1B            		/* Undefined Processing Undefined Instructions Mode 			 */
.set  MODE_SYS, 0x1F            		/* System Running Priviledged Operating System Tasks  Mode		 */
.set  I_BIT, 0x80               		/* when I bit is set, IRQ is disabled (program status registers) */
.set  F_BIT, 0x40               		/* when F bit is set, FIQ is disabled (program status registers) */

.text
.arm

.section .startup
.global	Reset_Handler
.global _vec_reset
.func _vec_reset

_vec_reset:

# Exception Vectors
.ifdef MCUNOSTART
_vectors:       b _vectors         
                ldr     PC, Undef_Addr
                ldr     PC, SWI_Addr
                ldr     PC, PAbt_Addr
                ldr     PC, DAbt_Addr
                nop							/* Reserved Vector */
                ldr     PC, IRQ_Addr
                ldr     PC, FIQ_Addr
.else
_vectors:       ldr     PC, Reset_Addr         
                ldr     PC, Undef_Addr
                ldr     PC, SWI_Addr
                ldr     PC, PAbt_Addr
                ldr     PC, DAbt_Addr
                nop							/* Reserved Vector */
                ldr     PC, IRQ_Addr
                ldr     PC, FIQ_Addr
.endif

Reset_Addr:     .word   Reset_Handler		/* defined in this module below  */
Undef_Addr:     .word   UNDEF_Routine		/* defined in irq.c  */
SWI_Addr:       .word   SWI_Routine			/* defined in irq.c  */
PAbt_Addr:      .word   PAbt_Routine		/* defined in irq.c  */
DAbt_Addr:      .word   DAbt_Routine		/* defined in irq.c  */
IRQ_Addr:       .word   IRQ_Routine			/* defined in irq.c  */
FIQ_Addr:       .word   FIQ_Routine			/* defined in irq.c  */

# Reset Handler

Reset_Handler:  

.if PLL_SETUP
                LDR     R0, =MMR_BASE_0
                MOV     R1, #0x01         
                STR     R1, [R0,#POWKEY1_OFFSET]          
                MOV     R1, #PLLCFG_Val      
                STR     R1, [R0,#POWCON_OFFSET]    
                MOV     R1, #0xF4
                STR     R1, [R0,#POWKEY2_OFFSET]
.endif

# Setup Stack for each mode 
                LDR     R0, =_stack_end 
 
#  Enter Undefined Instruction Mode and set its Stack Pointer 
                MSR     CPSR_c, #MODE_UND|I_BIT|F_BIT 
                MOV     SP, R0 
                SUB     R0, R0, #UND_STACK_SIZE 
 
#  Enter Abort Mode and set its Stack Pointer 
                MSR     CPSR_c, #MODE_ABT|I_BIT|F_BIT 
                MOV     SP, R0 
                SUB     R0, R0, #ABT_STACK_SIZE 
 
#  Enter FIQ Mode and set its Stack Pointer 
                MSR     CPSR_c, #MODE_FIQ|I_BIT|F_BIT 
                MOV     SP, R0 
                SUB     R0, R0, #FIQ_STACK_SIZE 
 
#  Enter IRQ Mode and set its Stack Pointer 
                MSR     CPSR_c, #MODE_IRQ|I_BIT|F_BIT 
                MOV     SP, R0 
                SUB     R0, R0, #IRQ_STACK_SIZE 
 
#  Enter Supervisor Mode and set its Stack Pointer 
                MSR     CPSR_c, #MODE_SVC|I_BIT|F_BIT 
                MOV     SP, R0 
                SUB     R0, R0, #SVC_STACK_SIZE
 
#  Enter User Mode and set its Stack Pointer 
                MSR     CPSR_c, #MODE_USR 
                MOV     SP, R0 
 
				/* copy initialized variables .data section  (Copy from ROM to RAM) */
                ldr     R1, =_etext
                ldr     R2, =_data
                ldr     R3, =_edata
1:        		cmp     R2, R3
                ldrlo   R0, [R1], #4
                strlo   R0, [R2], #4
                blo     1b

				/* Clear uninitialized variables .bss section (Zero init)  */
                mov     R0, #0
                ldr     R1, =_bss_start
                ldr     R2, =_bss_end
2:				cmp     R1, R2
                strlo   R0, [R1], #4
                blo     2b

				/* Enter the C code  */
                b       main

.endfunc
.end
