/********************************************************************/
/*                                                                  */
/*		Uart Funktionen zum Senden und Empfangen von Textfolgen		*/
/*		und 16bit,also short int ueber den uart.                    */
/*                                                                  */
/*		(C) Tilo Lutz, TiloLutz@gmx.de								*/
/*																	*/
/*		03.11.2006	Basisfunktionen von Marcus Moser uebernommen.   */
/*					Gesammte funktionen ueberarbeitet und ergaenzt.	*/
/*      24.12.2006  Ueberarbeitung der Variablen und Kommentare     */
/*		20.01.2007  Umschreiben in Objekt							*/
/*		30.01.2007  Bereinigung, so dass der Code verstaendlicher w.*/
/*																	*/
/*																	*/
/********************************************************************/

#include "uart.h"
#include "aduc.h"

// Uart Variablen
static volatile unsigned char uarttxbuffer[128];	// Senderingspeicher
static volatile unsigned char uartrxbuffer[128];	// Empfangsringspeicher
static volatile unsigned char uarttxbufferstatus;	// 0 = Keine Daten, 1 = Daten in Puffer, 2 = Puffer voll
static volatile unsigned char readuarttxbuffer;	// Lese ab hier Messwerte auf Senderingspeicher
static volatile unsigned char writeuarttxbuffer;	// Schreibe ab hier Messwerte in Senderingspeicher
static volatile unsigned char uartrxbufferstatus;	// 0 = Keine Daten, 1 = Daten in Puffer, 2 = Puffer voll
static volatile unsigned char readuartrxbuffer;	// Lese ab hier Messwerte auf Empfangsringspeicher
static volatile unsigned char writeuartrxbuffer;	// Schreibe ab hier Messwerte in Empfangsringspeicher
static volatile unsigned char uartirqstate;		// IRQ aus?


char UartPutChar (char ch) {
	char RETURN;
	if (uarttxbufferstatus!=2) { // Puffer ist nicht voll
		if (!uartirqstate) {
			COMIEN0 &= 0x01; // Deaktiviere TX Interrupt
		}
		uarttxbuffer[writeuarttxbuffer] = ch; // Schreibe Zeichen in Sendepuffer
		writeuarttxbuffer++;
		writeuarttxbuffer %= sizeof(uarttxbuffer);
		uarttxbufferstatus = 1;
		if (writeuarttxbuffer == readuarttxbuffer) { // Sendepuffer ist nun voll -> Gebe als Fehler 0 zurueck
			uarttxbufferstatus=2;
			RETURN = 0; // Puffer voll
		}
		else RETURN = 1; // Alles OK
		if (uartirqstate) {
			UartTXWait(); // Interrupt wird nicht ausgefhrt -> warte bis Byte gesendet
		}
		else { // Stosse Ubertragung von Hand an, da kein IRQ ausgeloest wird, wenn TX am Anfang leer ist
			COMIEN0 |= 0x02; // Aktiviere TX Interrupt
		}
	}
	else RETURN = 0; // Puffer voll
	return RETURN;
}

char UartGetChar(void) { // Lese Zeichen aus Empfangspuffer
	char RETURN = 0;
	if (uartirqstate) { // Keine IRQs aktiv, ueberpruefe selbst ob neues Byte da ist
		UartIrq();
	}
	if (uartrxbufferstatus!=0) { // Zeichen in Empfangsringspeicher
		RETURN = uartrxbuffer[readuartrxbuffer];
		readuartrxbuffer++; // Aendere Position im Empfangsringspeicher
		readuartrxbuffer %= sizeof(uartrxbuffer);
		if (uartrxbufferstatus==2) uartrxbufferstatus = 1;
		if (writeuarttxbuffer == readuarttxbuffer) { // Empfangspuffer ist nun leer, setzte auf 0 zurueck
			uartrxbufferstatus = 0;
		}
	}
	return RETURN; // Gebe 0 zurueck, wenn kein Zeichen empfangen wurde
}

char UartRXGet(void) { // Warte bis Zeichen empfangen wurde und gebe dieses aus
	char ReturnChar = 0;
	while (!ReturnChar) {
		ReturnChar = UartGetChar();
	}
	return ReturnChar;
}

void UartTXWait(void) {
	while (uarttxbufferstatus!=0) {
		UartIrq();
	} // Warte so lange bis alle Zeichen gesendet wurden
}

void UartIrq (void) {
	if (COMSTA0 & 0x01) { // Zeichen in RX Puffer
		if (uartrxbufferstatus!=2) { // Es ist noch Platz im Puffer
			uartrxbuffer[writeuartrxbuffer] = COMRX;
			writeuartrxbuffer++; // Aendere Position im Empfangsringspeicher
			writeuartrxbuffer %= sizeof(uartrxbuffer);;
			if (writeuartrxbuffer == readuartrxbuffer) { // Sendepuffer ist nun voll -> Gebe als Fehler 0 zurueck
				uartrxbufferstatus=2; // Puffer ist voll
			}
			else
				uartrxbufferstatus = 1; // Im Puffer liegen nun Daten
		}
		else { // Ueberschreibe aeltesten Wert in Puffer, damit das aelteste Zeichen verloren geht
			uartrxbuffer[writeuartrxbuffer] = COMRX;
			writeuartrxbuffer++; // Aendere Position im Empfangsringspeicher
			writeuartrxbuffer %= sizeof(uartrxbuffer);;
			readuartrxbuffer++; // Aendere Position im Empfangsringspeicher. Leseposition muss ebenfalls geaendert werden
			readuartrxbuffer %= sizeof(uartrxbuffer);;
		}
	}
	if (COMSTA0 & 0x40) { // TX Puffer leer.
		if (uarttxbufferstatus!=0) { // Zeichen in Senderingspeicher
			COMTX = uarttxbuffer[readuarttxbuffer]; // Sende Zeichen
			readuarttxbuffer++; // Aendere Position im Senderingspeicher
			readuarttxbuffer %= sizeof(uarttxbuffer);
			if (readuarttxbuffer==writeuarttxbuffer) { // Puffer ist nun leer
				COMIEN0 = 0x01; // Schalte TX-IRQ ab, Senderingspeicher leer.
				uarttxbufferstatus = 0; // Setzte Status = 0, also keine Daten mehr in Puffer
			}
			else
				uarttxbufferstatus = 1; // Puffer kann nicht mehr voll sein, da ein Element gesendet wurde
		}
		if (uarttxbufferstatus==0) { // Keine Zeichen in Puffer, Schalte IRQ ab.
			COMIEN0 = 0x01;
		}
	}

}


void UartIrqDisable(void) {	// Deaktiviert den Uart IRQ
	uartirqstate = 1;
	COMIEN0 = 0;
	IRQCLR = UART_BIT; // Deaktiviere UART IRQ
}


void UartIrqEnable(void) {	// Aktiviert den Uart IRQ
	COMIEN0  = 0x01; // Aktiviere RX-Full
	IRQEN |= UART_BIT; // Enable UART IRQ
	uartirqstate = 0;
}

void UartInit (int rate) { // Initialisiert den Uart
	char CD;
	unsigned char i;
	unsigned short int dl;
	unsigned short int M[4];
	unsigned short int N[4];

	CD = (POWCON & 0x07); // aktuelle Taktrate
	dl = (41780000 >> (CD+5)) / rate;
	GP1CON |= 0x0011; // Aktiviere uart
	COMCON0 = 0x080; // Erlaube setzen der Baudrate
	// Formel zur Baudratenbestimmung aus uC Handbuch S. 62
	COMDIV0 = (dl & 0x00FF); // Setze Lowbyte
	COMDIV1 = (dl & 0xFF00) >> 8; // Setze Highbyte
	COMDIV2 = 0;
	// Fraktalteiler
	// 1. Schritt: Berechne moegliche Werte fuer M und N
	for (i=0; i<4; i++) {
		M[i] = (41780000 >> (CD+5)) / rate / (dl-i);
		N[i] = ((41780000  / rate ) << (6-CD))  / (dl-i) - (M[i] << 11);
		if ((M[i] > 0) && (M[i] < 4)) { // M im gueltigen Bereich
			if ( M[i] == 4 ) { // Sonderregel fuer vier, M==0 entspricht *4,N
				M[i] = 0;
			}
			if ((N[i] > 0) && (N[i] < 2048)) { // Gueltiges Wertepaar gefunden
				COMDIV2 = 0x8000 + (M[i] << 11) + N[i]; // Aktiviere Fractal Divider wenn sinnvoll
				COMDIV0 = ((dl-i) & 0x00FF); // Setze Lowbyte
				COMDIV1 = ((dl-i) & 0xFF00) >> 8; // Setze Highbyte
			}
		}
	}

	COMCON0 = 0x003; // Verwende 8bit Woerter mit 1 Stopbit
	// Setze Variablen auf 0, Startbedingung
	readuarttxbuffer=0;
	readuartrxbuffer=0;
	writeuarttxbuffer=0;
	writeuartrxbuffer=0;
	uarttxbufferstatus=0;
	uartrxbufferstatus=0;
	uartirqstate=0;
	COMIEN0 = 0x01; // Aktiviere Interrupt fuer RX Full
	IRQEN |= UART_BIT; // Enable UART IRQ
	return;
}

void UartStop (void) { // Deaktiviert den uart	
	while (uarttxbufferstatus!=0) { // Warte bis TX-Puffer leer
		if (COMIEN0 == 0x01)
			break; // IRQ aus, Senden ist gar nicht mglich.
	}
	GP1CON ^= 0x0011; // Deaktiviere uart
	COMCON0 = 0x080; // Erlaube setzen der Baudrate
	COMDIV0 = 0; // Setze Baudrate auf 0
	COMDIV1 = 0; // Setze Baudrate auf 0
	COMCON0 = 0; // Reset Uart Einstellungen
	COMIEN0 = 0; // Schalte Interrupt aus
	IRQCLR = UART_BIT; // Deaktiviere UART IRQ
	return;
}


char *UartWrite (char *ptr, char wait) { // Sende String ueber uart, Wenn wait==1, wird gewartet, bis der String vollstaendig
		// gesendet werden konnte. Soll nicht ewartet werden und der TX-Puffer laeuft ueber, wird die Position des Zeichen im
		// String zurueckgegeben, die nicht mehr gesendet werden konnte
	char *origptr = ptr;
	char CharSent;
	while (*ptr!='\0') {
		CharSent = UartPutChar (*ptr); // Uebergebe einzelnes Zeichen an Sendefunktion
		ptr++;
		if (CharSent==0) { // Puffer ist voll. Warte oder gebe Position des letzten gesendeten Zeichen zurueck
			if (wait){
				while (uarttxbufferstatus==2) {
				} // Warte bis Platz in TX-Puffer ist
			}
			else
				return ptr; // Gebe Position des noch zu sendenden Zeichen aus
		}
	}
	return origptr; // Gebe Pointer auf String zurueck, wenn alles gesendet wurde.
}

void UartSendInt(int integer, char wait) { // Sende Messwert
	char stringint[12];
	UartIntToStr(integer, stringint); // Wandle Messwert in String um.
	UartWrite(stringint, wait);
}

char UartGetString(char *ptr, short int len, char lineend) { // Liest einen string von Uart
	short int pos = 0;
	while (pos<len-1) {
		*ptr = UartRXGet();
		if (lineend && *ptr==0xd) { // Zeilenende empfangen
			pos=len;
		}
		else {
			ptr++; // Springe zu naechstem Zeichen
			pos++; // Springe zu naechstem Zeichen
		}
	}
	*ptr = 0; // Terminiere String
	return 0;
}

int UartGetInt(void) { // Liesst eine Integerzahl ueber Uart. 1. Moeglichkeit: Integerzahl hat genau 12 Stellen.
		// 2. Moeglichkeit: Integerzahl wird mit Zeilenende abgeschlossen
	char BufferString[12];
	UartGetString(BufferString, 12, 1); // Empfange Integerzahl
	return UartStrToInt(BufferString);
}


static char *UartIntToStr(signed int zahl, char* string) {
	signed char i;
  string[11]='\0';                  // String Terminator
  if( zahl < 0 ) {                  // ist die Zahl negativ?
    string[0] = '-';              
    zahl = -zahl;
  }
  else {
	    string[0] = '+';
  }
 
  for(i=10; i>=1; i--) {
    string[i]=(zahl % 10) +'0';     // Modulo rechnen, dann den ASCII-Code von '0' addieren
    zahl /= 10;
  }
  // Fixme: Remove starting zeros
  return string;
}


// Diese Funktion wandelt einen String in Int um. Rueckgabewert ist Pointer auf Int
static int UartStrToInt(char *String) {
	const int TEST[10] = { 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000, 1000000000 };
	int val=0;
	unsigned char len=0;
	unsigned char neg=0;
	char *workString = String;
	while (*workString!=0) { // Stelle fest, wie lang der String ist.
		len++;
		workString++;
		}
	if (*String == '-') { // Verarbeite Vorzeichen
		neg=1;
		len--;
		String++;
	}
	if (*String == '+') { // Verarbeite Vorzeichen
		len--;
		String++;
	}
	while (len!=0) { // Zaehle Dezimalstellen zusammen.
		val += (*String-0x30) * TEST[len-1];
		len--;
		String++; // Springe zu naechstem Element
	}
	if (neg==1) val *= -1; // passe Vorzeichen an.
	return val;
}
