/********************************************************************/
/*                                                                  */
/*		Uart Funktionen zum Senden und Empfangen von Textfolgen		*/
/*		und 16bit,also short int ueber den uart.                    */
/*                                                                  */
/*		(C) Tilo Lutz, TiloLutz@gmx.de								*/
/*																	*/
/*		03.11.2006	Basisfunktionen von Marcus Moser uebernommen.   */
/*					Gesammte funktionen ueberarbeitet und ergaenzt.	*/
/*      24.12.2006  Ueberarbeitung der Variablen und Kommentare     */
/*		20.01.2007  Umschreiben in Objekt							*/
/*		30.01.2007  Bereinigung, so dass der Code verstaendlicher w.*/
/*																	*/
/*																	*/
/********************************************************************/

#include "uart.h"

// Uart Variablen
volatile unsigned char uarttxbuffer[128];	// Senderingspeicher
volatile unsigned char uartrxbuffer[128];	// Empfangsringspeicher
volatile unsigned char uarttxbufferstatus;	// 0 = Keine Daten, 1 = Daten in Puffer, 2 = Puffer voll
volatile unsigned char readuarttxbuffer;	// Lese ab hier Messwerte auf Senderingspeicher
volatile unsigned char writeuarttxbuffer;	// Schreibe ab hier Messwerte in Senderingspeicher
volatile unsigned char uartrxbufferstatus;	// 0 = Keine Daten, 1 = Daten in Puffer, 2 = Puffer voll
volatile unsigned char readuartrxbuffer;	// Lese ab hier Messwerte auf Empfangsringspeicher
volatile unsigned char writeuartrxbuffer;	// Schreibe ab hier Messwerte in Empfangsringspeicher
volatile unsigned char uartirqstate;		// IRQ aus?



void UartPutChar (char ch) {
	if (uartirqstate) { // IRQ ist aus! -> Warte bis gesendet werden kann.
		while ((COMSTA0 & 0x40)==0) {}; // Warte bis Puffer leer ist
		COMTX = ch;
	}
	else {
		while (uarttxbufferstatus==2) {}; // Warte bis Platz im Puffer ist
		COMIEN0 &= 0x01; // Deaktiviere TX Interrupt
		uarttxbuffer[writeuarttxbuffer] = ch; // Schreibe Zeichen in Sendepuffer
		writeuarttxbuffer++;
		writeuarttxbuffer %= sizeof(uarttxbuffer);
		uarttxbufferstatus = 1; //Mindestens ein Zeichen im Puffer
		if (writeuarttxbuffer == readuarttxbuffer) { // Sendepuffer ist nun voll -> Gebe als Fehler 0 zurueck
			uarttxbufferstatus=2;
		}
		COMIEN0 |= 0x02; // Aktiviere TX Interrupt
	}
}

char UartGetChar(void) { // Lese Zeichen aus Empfangspuffer
	char RETURN;
	if (uartirqstate) { // IRQ ist aus! -> Warte bis empfangen wurde.
		while ((COMSTA0 & 0x01)==0) {}; // Warte bis Zeichen empfangen wurde
		RETURN = COMRX;
	}
	else {
		while (uartrxbufferstatus==0) {}; // Noch nichts empfangen
		RETURN = uartrxbuffer[readuartrxbuffer];
		readuartrxbuffer++; // Aendere Position im Empfangsringspeicher
		readuartrxbuffer %= sizeof(uartrxbuffer);
		if (uartrxbufferstatus==2) uartrxbufferstatus = 1;
		if (writeuarttxbuffer == readuarttxbuffer) { // Empfangspuffer ist nun leer, setzte auf 0 zurueck
			uartrxbufferstatus = 0;
		}
	}
	return RETURN;
}


void UartIrq (void) {
	if (COMSTA0 & 0x01) { // Zeichen in RX Puffer
		if (uartrxbufferstatus!=2) { // Es ist noch Platz im Puffer
			uartrxbuffer[writeuartrxbuffer] = COMRX;
			writeuartrxbuffer++; // Aendere Position im Empfangsringspeicher
			writeuartrxbuffer %= sizeof(uartrxbuffer);;
			if (writeuartrxbuffer == readuartrxbuffer) { // Sendepuffer ist nun voll -> Gebe als Fehler 0 zurueck
				uartrxbufferstatus=2; // Puffer ist voll
			}
			else
				uartrxbufferstatus = 1; // Im Puffer liegen nun Daten
		}
		else { // Ueberschreibe aeltesten Wert in Puffer, damit das aelteste Zeichen verloren geht
			uartrxbuffer[writeuartrxbuffer] = COMRX;
			writeuartrxbuffer++; // Aendere Position im Empfangsringspeicher
			writeuartrxbuffer %= sizeof(uartrxbuffer);;
			readuartrxbuffer++; // Aendere Position im Empfangsringspeicher. Leseposition muss ebenfalls geaendert werden
			readuartrxbuffer %= sizeof(uartrxbuffer);;
		}
	}
	if (COMSTA0 & 0x40) { // TX Puffer leer.
		if (uarttxbufferstatus!=0) { // Zeichen in Senderingspeicher
			COMTX = uarttxbuffer[readuarttxbuffer]; // Sende Zeichen
			readuarttxbuffer++; // Aendere Position im Senderingspeicher
			readuarttxbuffer %= sizeof(uarttxbuffer);
			if (readuarttxbuffer==writeuarttxbuffer) { // Puffer ist nun leer
				COMIEN0 = 0x01; // Schalte TX-IRQ ab, Senderingspeicher leer.
				uarttxbufferstatus = 0; // Setzte Status = 0, also keine Daten mehr in Puffer
			}
			else
				uarttxbufferstatus = 1; // Puffer kann nicht mehr voll sein, da ein Element gesendet wurde
		}
		if (uarttxbufferstatus==0) { // Keine Zeichen in Puffer, Schalte IRQ ab.
			COMIEN0 = 0x01;
		}
	}

}




void UartIrqDisable(void) {	// Deaktiviert den Uart IRQ
	while (uarttxbufferstatus!=0) {}; // Warte bis Puffer gesendet wurde
	uartirqstate = 1;
	COMIEN0 = 0;
	IRQCLR = UART_BIT; // Deaktiviere UART IRQ
}


void UartIrqEnable(void) {	// Aktiviert den Uart IRQ
	COMIEN0  = 0x01; // Aktiviere RX-Full
	IRQEN |= UART_BIT; // Enable UART IRQ
	uartirqstate = 0;
}

void UartInit (int rate) { // Initialisiert den Uart
	unsigned char CD;
	unsigned char i;
	unsigned short int dl;
	unsigned short int M[4];
	unsigned short int N[4];

	CD = (POWCON & 0x07); // aktuelle Taktrate
	dl = (41780000 >> (CD+5)) / rate;
	GP1CON |= 0x0011; // Aktiviere uart
	COMCON0 = 0x080; // Erlaube setzen der Baudrate
	// Formel zur Baudratenbestimmung aus uC Handbuch S. 62
	COMDIV0 = (dl & 0x00FF); // Setze Lowbyte
	COMDIV1 = (dl & 0xFF00) >> 8; // Setze Highbyte
	COMDIV2 = 0;
	// Fraktalteiler
	// 1. Schritt: Berechne moegliche Werte fuer M und N
	for (i=0; i<4; i++) {
		M[i] = (41780000 >> (CD+5)) / rate / (dl-i);
		N[i] = ((41780000  / rate ) << (6-CD))  / (dl-i) - (M[i] << 11);
		if ((M[i] > 0) && (M[i] < 4)) { // M im gueltigen Bereich
			if ( M[i] == 4 ) { // Sonderregel fuer vier, M==0 entspricht *4,N
				M[i] = 0;
			}
			if ((N[i] > 0) && (N[i] < 2048)) { // Gueltiges Wertepaar gefunden
				COMDIV2 = 0x8000 + (M[i] << 11) + N[i]; // Aktiviere Fractal Divider wenn sinnvoll
				COMDIV0 = ((dl-i) & 0x00FF); // Setze Lowbyte
				COMDIV1 = ((dl-i) & 0xFF00) >> 8; // Setze Highbyte
			}
		}
	}

	COMCON0 = 0x003; // Verwende 8bit Woerter mit 1 Stopbit
	// Setze Variablen auf 0, Startbedingung
	readuarttxbuffer=0;
	readuartrxbuffer=0;
	writeuarttxbuffer=0;
	writeuartrxbuffer=0;
	uarttxbufferstatus=0;
	uartrxbufferstatus=0;
	uartirqstate=0;
	COMIEN0 = 0x01; // Aktiviere Interrupt fuer RX Full
	IRQEN |= UART_BIT; // Enable UART IRQ
	return;
}

void UartStop (void) { // Deaktiviert den uart	
	while (uarttxbufferstatus!=0) { // Warte bis TX-Puffer leer
		if (COMIEN0 == 0x01)
			break; // IRQ aus, Senden ist gar nicht mglich.
	}
	GP1CON ^= 0x0011; // Deaktiviere uart
	COMCON0 = 0x080; // Erlaube setzen der Baudrate
	COMDIV0 = 0; // Setze Baudrate auf 0
	COMDIV1 = 0; // Setze Baudrate auf 0
	COMCON0 = 0; // Reset Uart Einstellungen
	COMIEN0 = 0; // Schalte Interrupt aus
	IRQCLR = UART_BIT; // Deaktiviere UART IRQ
	return;
}
